/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.gui;

import jalview.datamodel.AlignmentI;
import jalview.datamodel.HiddenColumns;
import jalview.datamodel.SearchResultsI;
import jalview.datamodel.SequenceGroup;
import jalview.datamodel.SequenceI;
import jalview.renderer.ScaleRenderer;
import jalview.renderer.ScaleRenderer.ScaleMark;
import jalview.viewmodel.ViewportListenerI;
import jalview.viewmodel.ViewportRanges;

import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.image.BufferedImage;
import java.beans.PropertyChangeEvent;
import java.util.List;

import javax.swing.JComponent;

/**
 * DOCUMENT ME!
 * 
 * @author $author$
 * @version $Revision$
 */
public class SeqCanvas extends JComponent implements ViewportListenerI
{
  final FeatureRenderer fr;

  final SequenceRenderer seqRdr;

  BufferedImage img;

  Graphics2D gg;

  AlignViewport av;

  boolean fastPaint = false;

  int LABEL_WEST;

  int LABEL_EAST;

  int cursorX = 0;

  int cursorY = 0;

  int charHeight = 0;

  int charWidth = 0;

  boolean fastpainting = false;

  AnnotationPanel annotations;

  /**
   * Creates a new SeqCanvas object.
   * 
   * @param av
   *          DOCUMENT ME!
   */
  public SeqCanvas(AlignmentPanel ap)
  {
    this.av = ap.av;
    updateViewport();
    fr = new FeatureRenderer(ap);
    seqRdr = new SequenceRenderer(av);
    setLayout(new BorderLayout());
    PaintRefresher.Register(this, av.getSequenceSetId());
    setBackground(Color.white);

    av.getRanges().addPropertyChangeListener(this);
  }

  public SequenceRenderer getSequenceRenderer()
  {
    return seqRdr;
  }

  public FeatureRenderer getFeatureRenderer()
  {
    return fr;
  }

  private void updateViewport()
  {
    charHeight = av.getCharHeight();
    charWidth = av.getCharWidth();
  }

  /**
   * DOCUMENT ME!
   * 
   * @param g
   *          DOCUMENT ME!
   * @param startx
   *          DOCUMENT ME!
   * @param endx
   *          DOCUMENT ME!
   * @param ypos
   *          DOCUMENT ME!
   */
  private void drawNorthScale(Graphics g, int startx, int endx, int ypos)
  {
    updateViewport();
    for (ScaleMark mark : new ScaleRenderer().calculateMarks(av, startx,
            endx))
    {
      int mpos = mark.column; // (i - startx - 1)
      if (mpos < 0)
      {
        continue;
      }
      String mstring = mark.text;

      if (mark.major)
      {
        if (mstring != null)
        {
          g.drawString(mstring, mpos * charWidth, ypos - (charHeight / 2));
        }
        g.drawLine((mpos * charWidth) + (charWidth / 2),
                (ypos + 2) - (charHeight / 2),
                (mpos * charWidth) + (charWidth / 2), ypos - 2);
      }
    }
  }

  /**
   * DOCUMENT ME!
   * 
   * @param g
   *          DOCUMENT ME!
   * @param startx
   *          DOCUMENT ME!
   * @param endx
   *          DOCUMENT ME!
   * @param ypos
   *          DOCUMENT ME!
   */
  void drawWestScale(Graphics g, int startx, int endx, int ypos)
  {
    FontMetrics fm = getFontMetrics(av.getFont());
    ypos += charHeight;

    if (av.hasHiddenColumns())
    {
      startx = av.getAlignment().getHiddenColumns()
              .adjustForHiddenColumns(startx);
      endx = av.getAlignment().getHiddenColumns()
              .adjustForHiddenColumns(endx);
    }

    int maxwidth = av.getAlignment().getWidth();
    if (av.hasHiddenColumns())
    {
      maxwidth = av.getAlignment().getHiddenColumns()
              .findColumnPosition(maxwidth) - 1;
    }

    // WEST SCALE
    for (int i = 0; i < av.getAlignment().getHeight(); i++)
    {
      SequenceI seq = av.getAlignment().getSequenceAt(i);
      int index = startx;
      int value = -1;

      while (index < endx)
      {
        if (jalview.util.Comparison.isGap(seq.getCharAt(index)))
        {
          index++;

          continue;
        }

        value = av.getAlignment().getSequenceAt(i).findPosition(index);

        break;
      }

      if (value != -1)
      {
        int x = LABEL_WEST - fm.stringWidth(String.valueOf(value))
                - charWidth / 2;
        g.drawString(value + "", x,
                (ypos + (i * charHeight)) - (charHeight / 5));
      }
    }
  }

  /**
   * DOCUMENT ME!
   * 
   * @param g
   *          DOCUMENT ME!
   * @param startx
   *          DOCUMENT ME!
   * @param endx
   *          DOCUMENT ME!
   * @param ypos
   *          DOCUMENT ME!
   */
  void drawEastScale(Graphics g, int startx, int endx, int ypos)
  {
    ypos += charHeight;

    if (av.hasHiddenColumns())
    {
      endx = av.getAlignment().getHiddenColumns()
              .adjustForHiddenColumns(endx);
    }

    SequenceI seq;
    // EAST SCALE
    for (int i = 0; i < av.getAlignment().getHeight(); i++)
    {
      seq = av.getAlignment().getSequenceAt(i);
      int index = endx;
      int value = -1;

      while (index > startx)
      {
        if (jalview.util.Comparison.isGap(seq.getCharAt(index)))
        {
          index--;

          continue;
        }

        value = seq.findPosition(index);

        break;
      }

      if (value != -1)
      {
        g.drawString(String.valueOf(value), 0,
                (ypos + (i * charHeight)) - (charHeight / 5));
      }
    }
  }


  /**
   * need to make this thread safe move alignment rendering in response to
   * slider adjustment
   * 
   * @param horizontal
   *          shift along
   * @param vertical
   *          shift up or down in repaint
   */
  public void fastPaint(int horizontal, int vertical)
  {
    if (fastpainting || gg == null)
    {
      return;
    }
    fastpainting = true;
    fastPaint = true;
    updateViewport();

    ViewportRanges ranges = av.getRanges();
    int sr = ranges.getStartRes();
    int er = ranges.getEndRes();
    int ss = ranges.getStartSeq();
    int es = ranges.getEndSeq();
    int transX = 0;
    int transY = 0;

    gg.copyArea(horizontal * charWidth, vertical * charHeight,
            img.getWidth(), img.getHeight(), -horizontal * charWidth,
            -vertical * charHeight);

    if (horizontal > 0) // scrollbar pulled right, image to the left
    {
      transX = (er - sr - horizontal) * charWidth;
      sr = er - horizontal;
    }
    else if (horizontal < 0)
    {
      er = sr - horizontal;
    }
    else if (vertical > 0) // scroll down
    {
      ss = es - vertical;

      if (ss < ranges.getStartSeq())
      { // ie scrolling too fast, more than a page at a time
        ss = ranges.getStartSeq();
      }
      else
      {
        transY = img.getHeight() - ((vertical + 1) * charHeight);
      }
    }
    else if (vertical < 0)
    {
      es = ss - vertical;

      if (es > ranges.getEndSeq())
      {
        es = ranges.getEndSeq();
      }
    }

    gg.translate(transX, transY);
    drawPanel(gg, sr, er, ss, es, 0);
    gg.translate(-transX, -transY);

    repaint();
    fastpainting = false;
  }

  /**
   * Definitions of startx and endx (hopefully): SMJS This is what I'm working
   * towards! startx is the first residue (starting at 0) to display. endx is
   * the last residue to display (starting at 0). starty is the first sequence
   * to display (starting at 0). endy is the last sequence to display (starting
   * at 0). NOTE 1: The av limits are set in setFont in this class and in the
   * adjustment listener in SeqPanel when the scrollbars move.
   */

  // Set this to false to force a full panel paint
  @Override
  public void paintComponent(Graphics g)
  {
    super.paintComponent(g);

    updateViewport();

    ViewportRanges ranges = av.getRanges();

    int width = getWidth();
    int height = getHeight();

    width -= (width % charWidth);
    height -= (height % charHeight);

    // selectImage is the selection group outline image
    BufferedImage selectImage = drawSelectionGroup(
            ranges.getStartRes(), ranges.getEndRes(),
            ranges.getStartSeq(), ranges.getEndSeq());

    if ((img != null) && (fastPaint
            || (getVisibleRect().width != g.getClipBounds().width)
            || (getVisibleRect().height != g.getClipBounds().height)))
    {
      BufferedImage lcimg = buildLocalImage(selectImage);
      g.drawImage(lcimg, 0, 0, this);
      fastPaint = false;
    }
    else if ((width > 0) && (height > 0))
    {
      // img is a cached version of the last view we drew, if any
      // if we have no img or the size has changed, make a new one
      if (img == null || width != img.getWidth()
              || height != img.getHeight())
      {
        img = setupImage();
        if (img == null)
        {
          return;
        }
        gg = (Graphics2D) img.getGraphics();
        gg.setFont(av.getFont());
      }

      if (av.antiAlias)
      {
        gg.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                RenderingHints.VALUE_ANTIALIAS_ON);
      }

      gg.setColor(Color.white);
      gg.fillRect(0, 0, img.getWidth(), img.getHeight());

      if (av.getWrapAlignment())
      {
        drawWrappedPanel(gg, getWidth(), getHeight(), ranges.getStartRes());
      }
      else
      {
        drawPanel(gg, ranges.getStartRes(), ranges.getEndRes(),
                ranges.getStartSeq(), ranges.getEndSeq(), 0);
      }

      // lcimg is a local *copy* of img which we'll draw selectImage on top of
      BufferedImage lcimg = buildLocalImage(selectImage);
      g.drawImage(lcimg, 0, 0, this);
    }
  }

  /**
   * Draw an alignment panel for printing
   * 
   * @param g1
   *          Graphics object to draw with
   * @param startRes
   *          start residue of print area
   * @param endRes
   *          end residue of print area
   * @param startSeq
   *          start sequence of print area
   * @param endSeq
   *          end sequence of print area
   */
  public void drawPanelForPrinting(Graphics g1, int startRes, int endRes,
          int startSeq, int endSeq)
  {
    drawPanel(g1, startRes, endRes, startSeq, endSeq, 0);

    BufferedImage selectImage = drawSelectionGroup(startRes, endRes,
            startSeq, endSeq);
    if (selectImage != null)
    {
      ((Graphics2D) g1).setComposite(AlphaComposite
              .getInstance(AlphaComposite.SRC_OVER));
      g1.drawImage(selectImage, 0, 0, this);
    }
  }

  /**
   * Draw a wrapped alignment panel for printing
   * 
   * @param g
   *          Graphics object to draw with
   * @param canvasWidth
   *          width of drawing area
   * @param canvasHeight
   *          height of drawing area
   * @param startRes
   *          start residue of print area
   */
  public void drawWrappedPanelForPrinting(Graphics g, int canvasWidth,
          int canvasHeight, int startRes)
  {
    SequenceGroup group = av.getSelectionGroup();

    drawWrappedPanel(g, canvasWidth, canvasHeight, startRes);

    if (group != null)
    {
      BufferedImage selectImage = null;
      try
      {
        selectImage = new BufferedImage(canvasWidth, canvasHeight,
                BufferedImage.TYPE_INT_ARGB); // ARGB so alpha compositing works
      } catch (OutOfMemoryError er)
      {
        System.gc();
        System.err.println("Print image OutOfMemory Error.\n" + er);
        new OOMWarning("Creating wrapped alignment image for printing", er);
      }
      if (selectImage != null)
      {
        Graphics2D g2 = selectImage.createGraphics();
        setupSelectionGroup(g2, selectImage);
        drawWrappedSelection(g2, group, canvasWidth, canvasHeight,
                startRes);

        g2.setComposite(
                AlphaComposite.getInstance(AlphaComposite.SRC_OVER));
        g.drawImage(selectImage, 0, 0, this);
        g2.dispose();
      }
    }
  }

  /*
   * Make a local image by combining the cached image img
   * with any selection
   */
  private BufferedImage buildLocalImage(BufferedImage selectImage)
  {
    // clone the cached image
    BufferedImage lcimg = new BufferedImage(img.getWidth(), img.getHeight(),
            img.getType());
    Graphics2D g2d = lcimg.createGraphics();
    g2d.drawImage(img, 0, 0, null);

    // overlay selection group on lcimg
    if (selectImage != null)
    {
      g2d.setComposite(
              AlphaComposite.getInstance(AlphaComposite.SRC_OVER));
      g2d.drawImage(selectImage, 0, 0, this);
    }
    g2d.dispose();

    return lcimg;
  }

  /*
   * Set up a buffered image of the correct height and size for the sequence canvas
   */
  private BufferedImage setupImage()
  {
    BufferedImage lcimg = null;

    int width = getWidth();
    int height = getHeight();

    width -= (width % charWidth);
    height -= (height % charHeight);

    if ((width < 1) || (height < 1))
    {
      return null;
    }

    try
    {
      lcimg = new BufferedImage(width, height,
              BufferedImage.TYPE_INT_ARGB); // ARGB so alpha compositing works
    } catch (OutOfMemoryError er)
    {
      System.gc();
      System.err.println(
              "Group image OutOfMemory Redraw Error.\n" + er);
      new OOMWarning("Creating alignment image for display", er);

      return null;
    }

    return lcimg;
  }

  /**
   * DOCUMENT ME!
   * 
   * @param cwidth
   *          DOCUMENT ME!
   * 
   * @return DOCUMENT ME!
   */
  public int getWrappedCanvasWidth(int cwidth)
  {
    FontMetrics fm = getFontMetrics(av.getFont());

    LABEL_EAST = 0;
    LABEL_WEST = 0;

    if (av.getScaleRightWrapped())
    {
      LABEL_EAST = fm.stringWidth(getMask());
    }

    if (av.getScaleLeftWrapped())
    {
      LABEL_WEST = fm.stringWidth(getMask());
    }

    return (cwidth - LABEL_EAST - LABEL_WEST) / charWidth;
  }

  /**
   * Generates a string of zeroes.
   * 
   * @return String
   */
  String getMask()
  {
    String mask = "00";
    int maxWidth = 0;
    int tmp;
    for (int i = 0; i < av.getAlignment().getHeight(); i++)
    {
      tmp = av.getAlignment().getSequenceAt(i).getEnd();
      if (tmp > maxWidth)
      {
        maxWidth = tmp;
      }
    }

    for (int i = maxWidth; i > 0; i /= 10)
    {
      mask += "0";
    }
    return mask;
  }

  /**
   * DOCUMENT ME!
   * 
   * @param g
   *          DOCUMENT ME!
   * @param canvasWidth
   *          DOCUMENT ME!
   * @param canvasHeight
   *          DOCUMENT ME!
   * @param startRes
   *          DOCUMENT ME!
   */
  private void drawWrappedPanel(Graphics g, int canvasWidth,
          int canvasHeight, int startRes)
  {
    updateViewport();
    AlignmentI al = av.getAlignment();

    FontMetrics fm = getFontMetrics(av.getFont());

    LABEL_EAST = 0;
    LABEL_WEST = 0;

    if (av.getScaleRightWrapped())
    {
      LABEL_EAST = fm.stringWidth(getMask());
    }

    if (av.getScaleLeftWrapped())
    {
      LABEL_WEST = fm.stringWidth(getMask());
    }

    int hgap = charHeight;
    if (av.getScaleAboveWrapped())
    {
      hgap += charHeight;
    }

    int cWidth = (canvasWidth - LABEL_EAST - LABEL_WEST) / charWidth;
    int cHeight = av.getAlignment().getHeight() * charHeight;

    av.setWrappedWidth(cWidth);

    av.getRanges().setViewportStartAndWidth(startRes, cWidth);

    int endx;
    int ypos = hgap;
    int maxwidth = av.getAlignment().getWidth();

    if (av.hasHiddenColumns())
    {
      maxwidth = av.getAlignment().getHiddenColumns()
              .findColumnPosition(maxwidth);
    }

    while ((ypos <= canvasHeight) && (startRes < maxwidth))
    {
      endx = startRes + cWidth - 1;

      if (endx > maxwidth)
      {
        endx = maxwidth;
      }

      g.setFont(av.getFont());
      g.setColor(Color.black);

      if (av.getScaleLeftWrapped())
      {
        drawWestScale(g, startRes, endx, ypos);
      }

      if (av.getScaleRightWrapped())
      {
        g.translate(canvasWidth - LABEL_EAST, 0);
        drawEastScale(g, startRes, endx, ypos);
        g.translate(-(canvasWidth - LABEL_EAST), 0);
      }

      g.translate(LABEL_WEST, 0);

      if (av.getScaleAboveWrapped())
      {
        drawNorthScale(g, startRes, endx, ypos);
      }

      if (av.hasHiddenColumns() && av.getShowHiddenMarkers())
      {
        g.setColor(Color.blue);
        int res;
        HiddenColumns hidden = av.getAlignment().getHiddenColumns();
        List<Integer> positions = hidden.findHiddenRegionPositions();
        for (int pos : positions)
        {
          res = pos - startRes;

          if (res < 0 || res > endx - startRes)
          {
            continue;
          }

          gg.fillPolygon(
                  new int[]
                  { res * charWidth - charHeight / 4,
                      res * charWidth + charHeight / 4, res * charWidth },
                  new int[]
                  { ypos - (charHeight / 2), ypos - (charHeight / 2),
                      ypos - (charHeight / 2) + 8 },
                  3);

        }
      }

      // When printing we have an extra clipped region,
      // the Printable page which we need to account for here
      Shape clip = g.getClip();

      if (clip == null)
      {
        g.setClip(0, 0, cWidth * charWidth, canvasHeight);
      }
      else
      {
        g.setClip(0, (int) clip.getBounds().getY(), cWidth * charWidth,
                (int) clip.getBounds().getHeight());
      }

      drawPanel(g, startRes, endx, 0, al.getHeight() - 1, ypos);

      if (av.isShowAnnotation())
      {
        g.translate(0, cHeight + ypos + 3);
        if (annotations == null)
        {
          annotations = new AnnotationPanel(av);
        }

        annotations.renderer.drawComponent(annotations, av, g, -1, startRes,
                endx + 1);
        g.translate(0, -cHeight - ypos - 3);
      }
      g.setClip(clip);
      g.translate(-LABEL_WEST, 0);

      ypos += cHeight + getAnnotationHeight() + hgap;

      startRes += cWidth;
    }
  }

  /*
   * Draw a selection group over a wrapped alignment
   */
  private void drawWrappedSelection(Graphics2D g, SequenceGroup group,
          int canvasWidth,
          int canvasHeight, int startRes)
  {
    // height gap above each panel
    int hgap = charHeight;
    if (av.getScaleAboveWrapped())
    {
      hgap += charHeight;
    }

    int cWidth = (canvasWidth - LABEL_EAST - LABEL_WEST) / charWidth;
    int cHeight = av.getAlignment().getHeight() * charHeight;

    int startx = startRes;
    int endx;
    int ypos = hgap; // vertical offset
    int maxwidth = av.getAlignment().getWidth();

    if (av.hasHiddenColumns())
    {
      maxwidth = av.getAlignment().getHiddenColumns()
              .findColumnPosition(maxwidth);
    }

    // chop the wrapped alignment extent up into panel-sized blocks and treat
    // each block as if it were a block from an unwrapped alignment
    while ((ypos <= canvasHeight) && (startx < maxwidth))
    {
      // set end value to be start + width, or maxwidth, whichever is smaller
      endx = startx + cWidth - 1;

      if (endx > maxwidth)
      {
        endx = maxwidth;
      }

      g.translate(LABEL_WEST, 0);

      drawUnwrappedSelection(g, group, startx, endx, 0,
              av.getAlignment().getHeight() - 1,
              ypos);

      g.translate(-LABEL_WEST, 0);

      // update vertical offset
      ypos += cHeight + getAnnotationHeight() + hgap;

      // update horizontal offset
      startx += cWidth;
    }
  }

  int getAnnotationHeight()
  {
    if (!av.isShowAnnotation())
    {
      return 0;
    }

    if (annotations == null)
    {
      annotations = new AnnotationPanel(av);
    }

    return annotations.adjustPanelHeight();
  }

  /*
   * Draw an alignment panel for printing
   * 
   * @param g1
   *          Graphics object to draw with
   * @param startRes
   *          start residue of print area
   * @param endRes
   *          end residue of print area
   * @param startSeq
   *          start sequence of print area
   * @param endSeq
   *          end sequence of print area
   * @param offset
   *          vertical offset
   */
  private void drawPanel(Graphics g1, int startRes, int endRes,
          int startSeq, int endSeq, int offset)
  {
    updateViewport();
    if (!av.hasHiddenColumns())
    {
      draw(g1, startRes, endRes, startSeq, endSeq, offset);
    }
    else
    {
      int screenY = 0;
      int blockStart = startRes;
      int blockEnd = endRes;

      for (int[] region : av.getAlignment().getHiddenColumns()
              .getHiddenColumnsCopy())
      {
        int hideStart = region[0];
        int hideEnd = region[1];

        if (hideStart <= blockStart)
        {
          blockStart += (hideEnd - hideStart) + 1;
          continue;
        }

        blockEnd = hideStart - 1;

        g1.translate(screenY * charWidth, 0);

        draw(g1, blockStart, blockEnd, startSeq, endSeq, offset);

        if (av.getShowHiddenMarkers())
        {
          g1.setColor(Color.blue);

          g1.drawLine((blockEnd - blockStart + 1) * charWidth - 1,
                  0 + offset, (blockEnd - blockStart + 1) * charWidth - 1,
                  (endSeq - startSeq + 1) * charHeight + offset);
        }

        g1.translate(-screenY * charWidth, 0);
        screenY += blockEnd - blockStart + 1;
        blockStart = hideEnd + 1;

        if (screenY > (endRes - startRes))
        {
          // already rendered last block
          return;
        }
      }

      if (screenY <= (endRes - startRes))
      {
        // remaining visible region to render
        blockEnd = blockStart + (endRes - startRes) - screenY;
        g1.translate(screenY * charWidth, 0);
        draw(g1, blockStart, blockEnd, startSeq, endSeq, offset);

        g1.translate(-screenY * charWidth, 0);
      }
    }

  }

  private void draw(Graphics g, int startRes, int endRes, int startSeq,
          int endSeq, int offset)
  {
    g.setFont(av.getFont());
    seqRdr.prepare(g, av.isRenderGaps());

    SequenceI nextSeq;

    // / First draw the sequences
    // ///////////////////////////
    for (int i = startSeq; i <= endSeq; i++)
    {
      nextSeq = av.getAlignment().getSequenceAt(i);
      if (nextSeq == null)
      {
        // occasionally, a race condition occurs such that the alignment row is
        // empty
        continue;
      }
      seqRdr.drawSequence(nextSeq, av.getAlignment().findAllGroups(nextSeq),
              startRes, endRes, offset + ((i - startSeq) * charHeight));

      if (av.isShowSequenceFeatures())
      {
        fr.drawSequence(g, nextSeq, startRes, endRes,
                offset + ((i - startSeq) * charHeight), false);
      }

      // / Highlight search Results once all sequences have been drawn
      // ////////////////////////////////////////////////////////
      if (av.hasSearchResults())
      {
        int[] visibleResults = av.getSearchResults().getResults(nextSeq,
                startRes, endRes);
        if (visibleResults != null)
        {
          for (int r = 0; r < visibleResults.length; r += 2)
          {
            seqRdr.drawHighlightedText(nextSeq, visibleResults[r],
                    visibleResults[r + 1], (visibleResults[r] - startRes)
                            * charWidth, offset
                            + ((i - startSeq) * charHeight));
          }
        }
      }

      if (av.cursorMode && cursorY == i && cursorX >= startRes
              && cursorX <= endRes)
      {
        seqRdr.drawCursor(nextSeq, cursorX, (cursorX - startRes) * charWidth,
                offset + ((i - startSeq) * charHeight));
      }
    }

    if (av.getSelectionGroup() != null
            || av.getAlignment().getGroups().size() > 0)
    {
      drawGroupsBoundaries(g, startRes, endRes, startSeq, endSeq, offset);
    }

  }

  void drawGroupsBoundaries(Graphics g1, int startRes, int endRes,
          int startSeq, int endSeq, int offset)
  {
    Graphics2D g = (Graphics2D) g1;
    //
    // ///////////////////////////////////
    // Now outline any areas if necessary
    // ///////////////////////////////////

    SequenceGroup group = null;
    int groupIndex = -1;

    if (av.getAlignment().getGroups().size() > 0)
    {
      group = av.getAlignment().getGroups().get(0);
      groupIndex = 0;
    }

    if (group != null)
    {
      g.setStroke(new BasicStroke());
      g.setColor(group.getOutlineColour());
      
      do
      {
        drawPartialGroupOutline(g, group, startRes, endRes, startSeq,
                endSeq, offset);

        groupIndex++;

        g.setStroke(new BasicStroke());

        if (groupIndex >= av.getAlignment().getGroups().size())
        {
          break;
        }

        group = av.getAlignment().getGroups().get(groupIndex);

      } while (groupIndex < av.getAlignment().getGroups().size());

    }

  }


  /*
   * Draw the selection group as a separate image and overlay
   */
  private BufferedImage drawSelectionGroup(int startRes, int endRes,
          int startSeq, int endSeq)
  {
    // get a new image of the correct size
    BufferedImage selectionImage = setupImage();

    if (selectionImage == null)
    {
      return null;
    }

    SequenceGroup group = av.getSelectionGroup();
    if (group == null)
    {
      // nothing to draw
      return null;
    }

    // set up drawing colour
    Graphics2D g = (Graphics2D) selectionImage.getGraphics();

    setupSelectionGroup(g, selectionImage);

    if (!av.getWrapAlignment())
    {
      drawUnwrappedSelection(g, group, startRes, endRes, startSeq, endSeq,
              0);
    }
    else
    {
      drawWrappedSelection(g, group, getWidth(), getHeight(),
              av.getRanges().getStartRes());
    }

    g.dispose();
    return selectionImage;
  }

  /*
   * Set up graphics for selection group
   */
  private void setupSelectionGroup(Graphics2D g,
          BufferedImage selectionImage)
  {
    // set background to transparent
    g.setComposite(AlphaComposite.getInstance(AlphaComposite.CLEAR, 0.0f));
    g.fillRect(0, 0, selectionImage.getWidth(), selectionImage.getHeight());

    // set up foreground to draw red dashed line
    g.setComposite(AlphaComposite.Src);
    g.setStroke(new BasicStroke(1, BasicStroke.CAP_BUTT,
            BasicStroke.JOIN_ROUND, 3f, new float[]
    { 5f, 3f }, 0f));
    g.setColor(Color.RED);
  }

  /*
   * Draw a selection group over an unwrapped alignment
   * @param g graphics object to draw with
   * @param group selection group
   * @param startRes start residue of area to draw
   * @param endRes end residue of area to draw
   * @param startSeq start sequence of area to draw
   * @param endSeq end sequence of area to draw
   * @param offset vertical offset (used when called from wrapped alignment code)
   */
  private void drawUnwrappedSelection(Graphics2D g, SequenceGroup group,
          int startRes, int endRes, int startSeq, int endSeq, int offset)
  {
    if (!av.hasHiddenColumns())
    {
      drawPartialGroupOutline(g, group, startRes, endRes, startSeq, endSeq,
              offset);
    }
    else
    {
      // package into blocks of visible columns
      int screenY = 0;
      int blockStart = startRes;
      int blockEnd = endRes;

      for (int[] region : av.getAlignment().getHiddenColumns()
              .getHiddenColumnsCopy())
      {
        int hideStart = region[0];
        int hideEnd = region[1];

        if (hideStart <= blockStart)
        {
          blockStart += (hideEnd - hideStart) + 1;
          continue;
        }

        blockEnd = hideStart - 1;

        g.translate(screenY * charWidth, 0);
        drawPartialGroupOutline(g, group,
                blockStart, blockEnd, startSeq, endSeq, offset);

        g.translate(-screenY * charWidth, 0);
        screenY += blockEnd - blockStart + 1;
        blockStart = hideEnd + 1;

        if (screenY > (endRes - startRes))
        {
          // already rendered last block
          break;
        }
      }

      if (screenY <= (endRes - startRes))
      {
        // remaining visible region to render
        blockEnd = blockStart + (endRes - startRes) - screenY;
        g.translate(screenY * charWidth, 0);
        drawPartialGroupOutline(g, group,
                blockStart, blockEnd, startSeq, endSeq, offset);
        
        g.translate(-screenY * charWidth, 0);
      }
    }
  }

  /*
   * Draw the selection group as a separate image and overlay
   */
  private void drawPartialGroupOutline(Graphics2D g, SequenceGroup group,
          int startRes, int endRes, int startSeq, int endSeq,
          int verticalOffset)
  {
    int visWidth = (endRes - startRes + 1) * charWidth;

    int oldY = -1;
    int i = 0;
    boolean inGroup = false;
    int top = -1;
    int bottom = -1;

    int sx = -1;
    int sy = -1;
    int xwidth = -1;

    for (i = startSeq; i <= endSeq; i++)
    {
      // position of start residue of group relative to startRes, in pixels
      sx = (group.getStartRes() - startRes) * charWidth;

      // width of group in pixels
      xwidth = (((group.getEndRes() + 1) - group.getStartRes()) * charWidth)
              - 1;

      sy = verticalOffset + (i - startSeq) * charHeight;

      if (sx + xwidth < 0 || sx > visWidth)
      {
        continue;
      }

      if ((sx <= (endRes - startRes) * charWidth)
              && group.getSequences(null)
                      .contains(av.getAlignment().getSequenceAt(i)))
      {
        if ((bottom == -1) && !group.getSequences(null)
                .contains(av.getAlignment().getSequenceAt(i + 1)))
        {
          bottom = sy + charHeight;
        }

        if (!inGroup)
        {
          if (((top == -1) && (i == 0)) || !group.getSequences(null)
                  .contains(av.getAlignment().getSequenceAt(i - 1)))
          {
            top = sy;
          }

          oldY = sy;
          inGroup = true;
        }
      }
      else
      {
        if (inGroup)
        {
          // if start position is visible, draw vertical line to left of
          // group
          if (sx >= 0 && sx < visWidth)
          {
            g.drawLine(sx, oldY, sx, sy);
          }

          // if end position is visible, draw vertical line to right of
          // group
          if (sx + xwidth < visWidth)
          {
            g.drawLine(sx + xwidth, oldY, sx + xwidth, sy);
          }

          if (sx < 0)
          {
            xwidth += sx;
            sx = 0;
          }

          // don't let width extend beyond current block, or group extent
          // fixes JAL-2672
          if (sx + xwidth >= (endRes - startRes + 1) * charWidth)
          {
            xwidth = (endRes - startRes + 1) * charWidth - sx;
          }
          
          // draw horizontal line at top of group
          if (top != -1)
          {
            g.drawLine(sx, top, sx + xwidth, top);
            top = -1;
          }

          // draw horizontal line at bottom of group
          if (bottom != -1)
          {
            g.drawLine(sx, bottom, sx + xwidth, bottom);
            bottom = -1;
          }

          inGroup = false;
        }
      }
    }

    if (inGroup)
    {
      sy = verticalOffset + ((i - startSeq) * charHeight);
      if (sx >= 0 && sx < visWidth)
      {
        g.drawLine(sx, oldY, sx, sy);
      }

      if (sx + xwidth < visWidth)
      {
        g.drawLine(sx + xwidth, oldY, sx + xwidth, sy);
      }

      if (sx < 0)
      {
        xwidth += sx;
        sx = 0;
      }

      if (sx + xwidth > visWidth)
      {
        xwidth = visWidth;
      }
      else if (sx + xwidth >= (endRes - startRes + 1) * charWidth)
      {
        xwidth = (endRes - startRes + 1) * charWidth;
      }

      if (top != -1)
      {
        g.drawLine(sx, top, sx + xwidth, top);
        top = -1;
      }

      if (bottom != -1)
      {
        g.drawLine(sx, bottom - 1, sx + xwidth, bottom - 1);
        bottom = -1;
      }

      inGroup = false;
    }
  }
  
  /**
   * DOCUMENT ME!
   * 
   * @param results
   *          DOCUMENT ME!
   */
  public void highlightSearchResults(SearchResultsI results)
  {
    img = null;

    av.setSearchResults(results);

    repaint();
  }

  @Override
  public void propertyChange(PropertyChangeEvent evt)
  {
    String eventName = evt.getPropertyName();

    if (eventName.equals(SequenceGroup.SEQ_GROUP_CHANGED))
    {
      fastPaint = true;
      repaint();
    }
    else if (av.getWrapAlignment())
    {
      if (eventName.equals(ViewportRanges.STARTRES))
      {
        repaint();
      }
    }
    else
    {
      int scrollX = 0;
      if (eventName.equals(ViewportRanges.STARTRES))
      {
        // Make sure we're not trying to draw a panel
        // larger than the visible window
        ViewportRanges vpRanges = av.getRanges();
        scrollX = (int) evt.getNewValue() - (int) evt.getOldValue();
        int range = vpRanges.getEndRes() - vpRanges.getStartRes();
        if (scrollX > range)
        {
          scrollX = range;
        }
        else if (scrollX < -range)
        {
          scrollX = -range;
        }
      }

      // Both scrolling and resizing change viewport ranges: scrolling changes
      // both start and end points, but resize only changes end values.
      // Here we only want to fastpaint on a scroll, with resize using a normal
      // paint, so scroll events are identified as changes to the horizontal or
      // vertical start value.
      if (eventName.equals(ViewportRanges.STARTRES))
      {
        // scroll - startres and endres both change
        fastPaint(scrollX, 0);
      }
      else if (eventName.equals(ViewportRanges.STARTSEQ))
      {
        // scroll
        fastPaint(0, (int) evt.getNewValue() - (int) evt.getOldValue());
      }
    }
  }
}
